# Copyright (c) OpenMMLab. All rights reserved.
from .actionrecog_inferencer import ActionRecogInferencer
from .mmaction2_inferencer import MMAction2Inferencer

__all__ = ['ActionRecogInferencer', 'MMAction2Inferencer']
